{ This library is free software; you can redistribute it and/or modify it
  under the terms of the GNU Library General Public License as published by
  the Free Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public License
  for more details.

  You should have received a copy of the GNU Library General Public License
  along with this library; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

  ---------------------------------------------------------------------------

    Author : Luc DAVID Email: luckylazarus@free.fr
    2006 - 2009

    Major changes are indicated in the \Documentation\Changes.pdf file
    Last update 06.08.2009

  --------------------------------------------------------------------------- }

{ TSqlitePassRecordset}

constructor TSqlitePassRecordset.Create(Dataset: TSqlitePassDataset);
begin
  Create(Dataset.Database);
  FDataset := Dataset;
end;

constructor TSqlitePassRecordset.Create(Database: TSqlitePassDatabase);
begin
  inherited Create;
  FDataset := nil;
  FDatabase := Database;
  FActive := False;
  FPostponed := False;

  { Minimum Capacity is set to 100 records
  using a higher value when dealing with very large tables could improve speed }
  FRecordMemoryBlocCapacity := 100;

  {FRecordsMemoryBloc will be allocated in the GetRecords procedure,
   once we know the record size }
  FRecordsMemoryBlocs := TList.Create;
  FRecycledRecords := TList.Create;

  FAnsiStrings := TSqlitePassAnsiStringList.Create;
  FAnsiStrings.Capacity := FRecordsCapacity;
  FRecycledAnsiStrings := TList.Create;

  FWideStrings := TSqlitePassWideStringList.Create;
  FWideStrings.Capacity := FRecordsCapacity;
  FRecycledWideStrings := TList.Create;

  FMemoryStreams := TList.Create;
  FRecycledMemoryStreams := TList.Create;

  FActiveRecordOldValues := nil;
  FBlankRecordInitialization := nil;
  FNewLoadedRecordInitialization := nil;
  FCalcFieldsNullValuesInitBuffer := nil;
  FUnModifiedRecordInitialization := nil;
end;

destructor TSqlitePassRecordset.Destroy;
begin
  Close;
  FMemoryStreams.Free;
  FRecycledMemoryStreams.Free;
  FAnsiStrings.Free;
  FRecycledAnsiStrings.Free;
  FWideStrings.Free;
  FRecycledWideStrings.Free;
  FRecycledRecords.Free;
  FRecordsMemoryBlocs.Free;
  FDataset := nil;
  FDatabase := nil;
  inherited Destroy;
end;

{ --- Opening / Closing --- }

Procedure TSqlitePassRecordset.Open(Const Sql: String; TablesCount: Word = 0);
begin
Close;
Try
 FActive := True;
 FRowidTablesCount := TablesCount;
 GetRecords(Sql);
Except
 On EAbort do Close; { Safe call to free memory if needed }
    else begin
         Close; { Safe call to free memory if needed }
         FDatabase.DatabaseError.RaiseExceptionFmt(Msg3013, [CRLF, Sql], -1, FDatabase);
         end;
end;
end;

procedure TSqlitePassRecordset.Close;
begin
if FActive then
   try
   ClearMemory;
   finally
   FActive := False;
   end;
end;

procedure TSqlitePassRecordset.ClearMemory;
var
i: Integer;
Save_Cursor:TCursor;

begin
Try
   { Set the screen cursor }
   Save_Cursor := Screen.Cursor;
   Screen.Cursor := crHourglass;

   SetLength(FDataReader, 0);

   FCurrentRecordIndex := -1;
   FreeMem(FActiveRecordOldValues);
   FActiveRecordOldValues := nil;

   { Clear Records Pointers }
   SetFRecordsCapacity(0);
   FRecordsCount := 0;
   FRecordsUsedInMemoryBloc := FRecordMemoryBlocCapacity;
   FRecycledRecords.Clear;

   { Clear FRecordsMemoryBlocs }
   For i := 0 to Pred(FRecordsMemoryBlocs.Count)
       do FreeMem(FRecordsMemoryBlocs[i]);
   FRecordsMemoryBlocs.Clear;

   { Blank Record Buffer }
   FreeMem(FBlankRecordInitialization);
   FBlankRecordInitialization := nil;

   { Blank Record Buffer + Null value flags set to FALSE }
   FreeMem(FNewLoadedRecordInitialization);
   FNewLoadedRecordInitialization := nil;

   { UnModified Value Buffer }
   FreeMem(FUnModifiedRecordInitialization);
   FUnModifiedRecordInitialization := nil;

   { Clear Strings }
   FAnsiStrings.Clear;
   FRecycledAnsiStrings.Clear;

   { Clear WideStrings }
   FWideStrings.Clear;
   FRecycledWideStrings.Clear;

   { Clear Blobs and clear Blobs List pointer }
   For i := 0 to Pred(FMemoryStreams.Count)
       do TMemoryStream(FMemoryStreams[i]).Free;
   FMemoryStreams.Clear;
   FRecycledMemoryStreams.Clear;
   
   { Calc and Lookup null values}
   FreeMem(FCalcFieldsNullValuesInitBuffer);
   FCalcFieldsNullValuesInitBuffer := nil;
   {  }
   Finalize(FInternalFieldsOffset);
finally
Screen.Cursor := Save_Cursor;
end;
end;

{ Returns the active record index : the record that is currently used by dataset user }
function TSqlitePassRecordset.GetActiveRecordIndex: Integer;
begin
 if Assigned(FDataset)
    then Result := FDataset.GetActiveRecord
    else Result := FCurrentRecordIndex;
end;

{ Result : the memory size needed to store a given Datatype in the internal record buffer }
function TSqlitePassRecordset.GetFieldSize(DataType: TFieldType): Integer;
begin
  case DataType of
       ftInteger, ftSmallInt, ftWord, ftDate, ftTime,
       ftString, ftFixedChar, ftWideString: Result := SizeOf(Integer);
       ftDateTime, ftFloat, ftBCD, ftCurrency: Result := SizeOf(Double);
       ftBoolean: Result := SizeOf(WordBool);
       ftBlob, ftMemo, ftGraphic, ftFmtMemo, ftTypedBinary, ftBytes, ftVarBytes
       {$IFDEF HasftWideMemo}, ftWideMemo {$ENDIF}: Result := SizeOf(Integer);
       ftAutoInc, ftLargeint: Result := SizeOf(Int64);
       else Result := SizeOf(PRecBuffer);
       end;
end;


{ Calculates the size needed to store fields values, including calculated and lookup fields
  and others information in a record
  Returns the number of fields including calculated and lookup }
function TSqlitePassRecordset.SetFRecordSize: Integer;
var
i, j: Integer;
begin
  FCalcFieldsCount := 0;
  SetLength(FInternalFieldsOffset, 0);

  FBookMarkDataOffset := SizeOf(TBookMarkFlag);
  FRecordStateOffset := FBookMarkDataOffset + SizeOf(Cardinal);
  FRowIdsOffset :=  FRecordStateOffset + SizeOf(TSqlitePassRecordState);
  FFieldsValuesOffset :=  FRowIdsOffset + (FRowidTablesCount * SizeOf(Int64));
  FRecordSize := FFieldsValuesOffset;

  if Assigned(FDataset)
     then begin
          { Set the Fields offset to the maximum length. We will downsize it later }
          SetLength(FInternalFieldsOffset, FDataset.FieldDefs.Count + FDataset.Fields.Count);

          { fkData Fields }
          for i := 0 to Pred(FDataset.FieldDefs.Count)
              do begin
                 FInternalFieldsOffset[i] := FRecordSize;
                 Inc(FRecordSize, GetFieldSize(FDataset.FieldDefs[i].DataType));
                 end;
          { Other fields Types (Calculated or lookup) }
          j := FDataset.FieldDefs.Count;
          for i := 0 to Pred(FDataset.Fields.Count)
              do if FDataset.Fields[i].FieldKind <> fkData then
                 begin
                 FInternalFieldsOffset[j] := FRecordSize;
                 Inc(FRecordSize, GetFieldSize(FDataset.Fields[i].DataType));
                 Inc(j);
                 end;
          { Returns the number of fields, including calc or lookup fields }
          Result := j;
          { Initialize a memory bloc with null values. It will be used to Clear
            the Calc or Lookup fields content }
          FCalcFieldsCount := j-FDataset.FieldDefs.Count;
          if FCalcFieldsCount > 0 then
             begin
             GetMem(FCalcFieldsNullValuesInitBuffer, FCalcFieldsCount);
             for i := 0 to Pred(FCalcFieldsCount)
                do PByteBool(FCalcFieldsNullValuesInitBuffer + (i * SizeOf(ByteBool)))^ := True;
             end;

          { Set the definitive Fields Offset array size }
          SetLength(FInternalFieldsOffset, j);
         end
     { No TDataset.fields are defined so we use Strings to store any values (stored in FAnsiStrings array) }
     else begin
          SetLength(FInternalFieldsOffset, FDataFieldsCount);
          for i := 0 to Pred(FDataFieldsCount) do
              begin
              Inc(FRecordSize, SizeOf(Integer));
              FInternalFieldsOffset[i] := FRecordSize;
              end;
          Result := FDataFieldsCount;
          end;

  { Null values boolean flag buffer }
  FFieldsNullValuesOffset := FRecordSize;
  Inc(FRecordSize, SizeOf(ByteBool)* Result);

  { Modified values boolean flag buffer }
  FFieldsModifiedValuesOffset := FRecordSize;
  GetMem(FUnModifiedRecordInitialization, SizeOf(ByteBool)* Result);
  Inc(FRecordSize, SizeOf(ByteBool)* Result);

  { Prepare a memory bloc with 'blank' values. This memory bloc will be used to initialize
    each new record in the AddRecord function }
  GetMem(FBlankRecordInitialization, FRecordSize);

  { Prepare a memory bloc with 'blank' values and Null value flags to FALSE.
    This memory bloc will be used to initialize each new record in the _AddRecord function }
  GetMem(FNewLoadedRecordInitialization, FRecordSize);

  { Set Record state }
//  TSqlitePassRecordState((FBlankRecordInitialization + FRecordStateOffset)^) := rsUnModified;

  { Initialize Null values Flags to TRUE or FALSE and Modified Flags to FALSE}
  for j := 0 to Pred(Result) do
      begin
      i := j * SizeOf(ByteBool);
      PByteBool(FBlankRecordInitialization + FFieldsNullValuesOffset + i)^ := True;
      PByteBool(FBlankRecordInitialization + FFieldsModifiedValuesOffset + i)^ := False;
      PByteBool(FNewLoadedRecordInitialization + FFieldsNullValuesOffset + i)^ := False;
      PByteBool(FNewLoadedRecordInitialization + FFieldsModifiedValuesOffset + i)^ := False;
      PByteBool(FUnModifiedRecordInitialization + i)^ := False;
      end;
end;

procedure TSqlitePassRecordset.PrepareDataReader(FDataset: TDataset);
var
i: Integer;
begin
SetLength(FDataReader,0);

if Assigned(FDataset) then
   begin
   SetLength(FDataReader, FColumnsCount);

   { RowId Columns }
   for i := 0 to Pred(FRowIdTablesCount) do
   with FDataReader[i] do
        begin
        Load := @SqliteValueToInt64;
        BufferSize := SizeOf(Int64);
        end;

   { Dataset Fields Columns }
   for i := FRowIdTablesCount to Pred(FColumnsCount) do
   with FDataReader[i] do
        begin

        { Safe assignation to a dummy procedure }
        Load := @SqliteValueToDummy;

        Case FDataset.FieldDefs[i - FRowIdTablesCount].DataType of

          ftBoolean:
          begin
          BufferSize := SizeOf(WordBool);
          Case FDatabase.DatatypeOptions.FBooleanStorage of
               asInteger : Load := @SqliteValueToWordBool;
               asText    : Load := @SqliteValueTextToWordBool;
               end;
          end;

          ftSmallint, ftInteger, ftWord:
          begin
          Load := @SqliteValueToInteger;
          BufferSize := SizeOf(Integer);
          end;

          ftAutoInc, ftLargeint:
          begin
          Load := @SqliteValueToInt64;
          BufferSize := SizeOf(Int64);
          end;

          ftFloat, ftCurrency, ftBCD:
          begin
          Load := @SqliteValueToDouble;
          BufferSize := SizeOf(Double);
          end;

          ftDate:
          begin
          BufferSize := SizeOf(Integer);
          Case FDatabase.DatatypeOptions.DateStorage of
               asInteger : Load := @SqliteValueToInteger;
               asText    : Load := @SqliteValueTextToDate;
               end;
          end;

          ftTime:
          begin
          BufferSize := SizeOf(Integer);
          Case FDatabase.DatatypeOptions.TimeStorage of
               asInteger : Load := @SqliteValueToInteger;
               asText    : Load := @SqliteValueTextToTime;
               end;
          end;

          ftDateTime:
          begin
          BufferSize := SizeOf(Double);
          Case FDatabase.DatatypeOptions.DateTimeStorage of
               dtsDateTime : Load := @SqliteValueToDouble;
               dtsText     : Load := @SqliteValueDateTimeTextToDouble;
               dtsJulian   : Load := @SqliteValueJulianDateTimeToDouble;
               dtsUnix     : Load := @SqliteValueUnixDateTimeToDouble;
               dtsMac      : Load := @SqliteValueMacDateTimeToDouble;
               end;
          end;

          ftString, ftFixedChar:
          begin
          Load := @SqliteValueToAnsiString;
          BufferSize := SizeOf(Integer);
          end;

          ftWideString:
          begin
          Load := @SqliteValueToWideString;
          BufferSize := SizeOf(Integer);
          end;


          ftBlob, ftMemo, ftGraphic, ftFmtMemo, ftTypedBinary, ftBytes, ftVarBytes
         {$IFDEF HasftWideMemo}, ftWideMemo {$ENDIF} :
          begin
          Load := @SqliteValueToBlob;
          BufferSize := SizeOf(Integer);
          end;

          else begin
               Load := @SqliteValueToAnsiString;
               BufferSize := SizeOf(Integer);
               end;
          end; { Case}
       end; { For i := 0 to ...}
       end else { Not Assigned(FDataset) }
           begin
           SetLength(FDataReader, FDataFieldsCount);
           For i := 0 to Pred(FDataFieldsCount) do
               begin
               if FDatabase.FSystemEncoding = sysUTF8
                  then FDataReader[i].Load := @SqliteValueToAnsiString
                  else FDataReader[i].Load := @SqliteValueToWideString;

               FDataReader[i].BufferSize := SizeOf(Integer);
               end;
           end;
end;


procedure TSqlitePassRecordset.GetRecords(Const Sql: String);
var
Accept, IsNull: Boolean;
PreparedStmt: Pointer;
Buffer: PRecBuffer;

ColumnIndex,
RowIndex, RowLowerLimit, RowUpperLimit,
ErrorCode: Integer;
Save_Cursor:TCursor;
begin
Try

{ Set the screen cursor }
Save_Cursor := Screen.Cursor;
Screen.Cursor := crHourglass;

{ Clear Memory pointers, blocs...etc }
ClearMemory;

{ Prepare the SQL statement }
PreparedStmt := nil;
FDatabase.Engine.PrepareStmt(PreparedStmt, Sql);
{FColumnsCount is made of :
 n columns to store each table rowid + n columns for data }
FColumnsCount := SqliteDbv3_column_count(PreparedStmt);

{ Sometimes there is no data at all }
if FColumnsCount = 0
   then Abort;

{ Start Chrono }
FDatabase.Engine.StartChrono;

{ Get the Datafields Count. }
FDataFieldsCount := FColumnsCount - FRowIdTablesCount;

{ Set the record size and get the number of fields including calculated and lookup }
FFieldsCount := SetFRecordSize;

{ Set the Data Reader and Translator }
PrepareDataReader(FDataset);

{ Set the Lower and Upper Limit to retrieve only the rows requested }
RowIndex := 0;
RowLowerLimit := 1;
RowUpperLimit := Pred(MaxLongInt);

if Assigned(FDataset) then
  begin
  FRecordMemoryBlocCapacity := FDataset.RecordsCacheCapacity;
  FPostPoned := (FDataset.FWriteMode = wmPostponed);
  InitInternalFieldsInfos(FDataset.Fields);
  if FDataset.Filtered then
     begin
     RowLowerLimit := Max(1, FDataset.FilterRecordLowerLimit);
     if FDataset.FilterRecordUpperLimit <> 0
        then RowUpperLimit := Succ(FDataset.FilterRecordUpperLimit);
     InitInternalFilters(FDataset.Fields);
     end;
  end;

{ We step through recordset until we reach LowerLimit or we get an error }
repeat
  ErrorCode :=SqliteDbv3_Step(PreparedStmt);
  Inc(RowIndex);
until (ErrorCode <> SQLITE_ROW) or (RowIndex = RowLowerLimit);

{ Load the records : Proceed with every records until we reach UpperLimit or we get an error }
While (ErrorCode = SQLITE_ROW) and (RowIndex < RowUpperLimit) do
  begin
  { Update Records Count and Capacity }
  _AddRecord;

  { Initialize the new record with 'blank' values and Set Null flags to FALSE }
  Move(FNewLoadedRecordInitialization^, FRecords[FRecordsCount]^, FRecordSize);

  { Set the buffer to the new record }
  Buffer := FRecords[FRecordsCount];

  { Move the pointer to the first RowId or field Data buffer address }
  Inc(Buffer, FRowIdsOffset);

  { We get the RowId values for the current data row }
  for ColumnIndex := 0 to Pred(FRowIdTablesCount) do
      With FDataReader[ColumnIndex] do
      begin
      Load(Self, Buffer, PreparedStmt, ColumnIndex);
      Inc(Buffer, BufferSize);
      end;

  { We get the fields values for the current data row }
  for ColumnIndex := FRowIdTablesCount to Pred(FColumnsCount) do
      begin
      Try
      IsNull := SqliteDbv3_column_type(PreparedStmt, ColumnIndex) = SQLITE_NULL;
      With FDataReader[ColumnIndex] do
           begin
           if IsNull
              then SetFieldNullValue(FRecordsCount, ColumnIndex - FRowIdTablesCount, True)
              else Load(Self, Buffer, PreparedStmt, ColumnIndex);
           Inc(Buffer, BufferSize);
           end;
      except
      { Exception while retreiving data are trapped here
        so we can continu with the next available value }
      On E: Exception do;
      end;
      end;

  Inc(RowIndex);
  { Post the new record before deleting }
  Inc(FRecordsCount);
  Inc(FRecordsUsedInMemoryBloc);

  { If needed, we send filter event now }
  if Assigned(FDataset) then
     begin
     { dsFilter is set only when Dataset is Filtered and OnFilterRecord is Assigned }
     if FDataset.State = dsFilter then
        begin
        { Checks if record is accepted by the filter(s) }
        FCurrentRecordIndex := Pred(FRecordsCount);
        Accept := FDataset.FilterAcceptRecord(FCurrentRecordIndex);
        If Accept and Assigned(FDataset.OnFilterRecord)
           then FDataset.OnFilterRecord(FDataset, Accept);
        If Not Accept then
           begin
           DeleteRecord(FCurrentRecordIndex);
           Dec(RowIndex);
           end;
        end;
     end;

  ErrorCode :=SqliteDbv3_Step(PreparedStmt);
end; { While }

// TODO (FStatementTimeout=0) or ((FStatementTimeout>0) and (integer(GetTickCount)>=t));

{ Move to the first record }
FCurrentRecordIndex := -1;
FOldActiveRecordIndex := -1;

{ Create a utility record to be able to restore a record modified by user to
  it previous state if needed }
FActiveRecordOldValues := AllocMem(FRecordSize);

finally
Screen.Cursor := Save_Cursor;
{ Free the prepared Statement }
FDatabase.Engine.UnPrepareStmt(PreparedStmt);
{ Stop Chrono }
FDatabase.Engine.StopChrono;
end;
end;



{ functions to translate a SQLite value to its internal storage value in memory }

procedure SqliteValueToDummy
          (Const FRecordset: TSqlitePassRecordset;
           Const Buffer: PRecBuffer;
           Const PreparedStmt: Pointer;
           Const ColumnIndex: Integer);{$IFDEF HasInline} inline; {$ENDIF}
begin
{ Just a empty procedure to ensure DataReader procedure pointer is always assigned }
end;

procedure SqliteValueToWordBool
          (Const FRecordset: TSqlitePassRecordset;
           Const Buffer: PRecBuffer;
           Const PreparedStmt: Pointer;
           Const ColumnIndex: Integer);{$IFDEF HasInline} inline; {$ENDIF}
var
BooleanValue: WordBool;
begin
BooleanValue := SqliteDbv3_column_int(PreparedStmt,ColumnIndex) <> 0;
Move(BooleanValue, Buffer^, SizeOf(WordBool));
end;

procedure SqliteValueTextToWordBool
          (Const FRecordset: TSqlitePassRecordset;
           Const Buffer: PRecBuffer;
           Const PreparedStmt: Pointer;
           Const ColumnIndex: Integer);{$IFDEF HasInline} inline; {$ENDIF}
var
BooleanValue: WordBool;
begin
FRecordset.FDatabase.FTranslator.BooleanTextToWordBool(SqliteDbv3_column_text(PreparedStmt,ColumnIndex),BooleanValue);
Move(BooleanValue, Buffer^, SizeOf(WordBool));
end;


procedure SqliteValueToInteger(Const FRecordset: TSqlitePassRecordset;
                               Const Buffer: PRecBuffer;
                               Const PreparedStmt: Pointer;
                               Const ColumnIndex: Integer);{$IFDEF HasInline} inline; {$ENDIF}
var
IntegerValue: Integer;
begin
    IntegerValue := SqliteDbv3_column_int(PreparedStmt,ColumnIndex);
    Move(IntegerValue, Buffer^, SizeOf(Integer));
end;

procedure SqliteValueToInt64(Const FRecordset: TSqlitePassRecordset;
                             Const Buffer: PRecBuffer;
                             Const PreparedStmt: Pointer;
                             Const ColumnIndex: Integer);{$IFDEF HasInline} inline; {$ENDIF}
var
Int64Value: Int64;
begin
  Int64Value := SqliteDbv3_column_int64(PreparedStmt,ColumnIndex);
  Move(Int64Value, Buffer^, SizeOf(Int64));
end;

procedure SqliteValueToDouble
          (Const FRecordset: TSqlitePassRecordset;
           Const Buffer: PRecBuffer;
           Const PreparedStmt: Pointer;
           Const ColumnIndex: Integer);{$IFDEF HasInline} inline; {$ENDIF}
var
DoubleValue: Double;
begin
  DoubleValue := SqliteDbv3_column_double(PreparedStmt,ColumnIndex);
  Move(DoubleValue, Buffer^, SizeOf(Double));
end;

procedure SqliteValueTextToDate
          (Const FRecordset: TSqlitePassRecordset;
           Const Buffer: PRecBuffer;
           Const PreparedStmt: Pointer;
           Const ColumnIndex: Integer);{$IFDEF HasInline} inline; {$ENDIF}
var
IntegerValue: Integer;
begin
  FRecordset.FDatabase.FTranslator.DateTextToInteger(SqliteDbv3_column_text(PreparedStmt,ColumnIndex),IntegerValue);
  Move(IntegerValue, Buffer^, SizeOf(Integer));
end;

procedure SqliteValueTextToTime
          (Const FRecordset: TSqlitePassRecordset;
           Const Buffer: PRecBuffer;
           Const PreparedStmt: Pointer;
           Const ColumnIndex: Integer);{$IFDEF HasInline} inline; {$ENDIF}
var
IntegerValue: Integer;
begin
  FRecordset.FDatabase.FTranslator.TimeTextToInteger(SqliteDbv3_column_text(PreparedStmt,ColumnIndex),IntegerValue);
  Move(IntegerValue, Buffer^, SizeOf(Integer));
end;


procedure SqliteValueJulianDateTimeToDouble
          (Const FRecordset: TSqlitePassRecordset;
           Const Buffer: PRecBuffer;
           Const PreparedStmt: Pointer;
           Const ColumnIndex: Integer);{$IFDEF HasInline} inline; {$ENDIF}
var
DoubleValue: Double;
begin
  FRecordset.FDatabase.FTranslator.JulianDateTimeToDouble(SqliteDbv3_column_double(PreparedStmt,ColumnIndex),DoubleValue);
  Move(DoubleValue, Buffer^, SizeOf(Double));
end;

procedure SqliteValueDateTimeTextToDouble
          (Const FRecordset: TSqlitePassRecordset;
           Const Buffer: PRecBuffer;
           Const PreparedStmt: Pointer;
           Const ColumnIndex: Integer);{$IFDEF HasInline} inline; {$ENDIF}
var
DoubleValue: Double;
begin
  FRecordset.FDatabase.FTranslator.DateTimeTextToDouble(SqliteDbv3_column_text(PreparedStmt,ColumnIndex),DoubleValue);
  Move(DoubleValue, Buffer^, SizeOf(Double));
end;

procedure SqliteValueUnixDateTimeToDouble
          (Const FRecordset: TSqlitePassRecordset;
           Const Buffer: PRecBuffer;
           Const PreparedStmt: Pointer;
           Const ColumnIndex: Integer);{$IFDEF HasInline} inline; {$ENDIF}
var
DoubleValue: Double;
begin
  FRecordset.FDatabase.FTranslator.UnixDateTimeToDouble(SqliteDbv3_column_int64(PreparedStmt,ColumnIndex),DoubleValue);
  Move(DoubleValue, Buffer^, SizeOf(Double));
end;

procedure SqliteValueMacDateTimeToDouble
          (Const FRecordset: TSqlitePassRecordset;
           Const Buffer: PRecBuffer;
           Const PreparedStmt: Pointer;
           Const ColumnIndex: Integer);{$IFDEF HasInline} inline; {$ENDIF}
var
DoubleValue: Double;
begin
  FRecordset.FDatabase.FTranslator.MacDateTimeToDouble(SqliteDbv3_column_int64(PreparedStmt,ColumnIndex),DoubleValue);
  Move(DoubleValue, Buffer^, SizeOf(Double));
end;

procedure SqliteValueToAnsiString(Const FRecordset: TSqlitePassRecordset;
                                  Const Buffer: PRecBuffer;
                                  Const PreparedStmt: Pointer;
                                  Const ColumnIndex: Integer);{$IFDEF HasInline} inline; {$ENDIF}
begin
 FRecordset.AddAnsiString(Buffer, SqliteDbv3_column_text(PreparedStmt, ColumnIndex));
end;

procedure SqliteValueToWideString(Const FRecordset: TSqlitePassRecordset;
                                  Const Buffer: PRecBuffer;
                                  Const PreparedStmt: Pointer;
                                  Const ColumnIndex: Integer);{$IFDEF HasInline} inline; {$ENDIF}
begin
 FRecordset.AddWideString(Buffer, SqliteDbv3_column_text16(PreparedStmt, ColumnIndex));
end;

procedure SqliteValueToBlob(Const FRecordset: TSqlitePassRecordset;
                            Const Buffer: PRecBuffer;
                            Const PreparedStmt: Pointer;
                            Const ColumnIndex: Integer);{$IFDEF HasInline} inline; {$ENDIF}
var
BlobStream: TMemoryStream;
BlobSize: Cardinal;
begin
BlobSize := SqliteDbv3_column_bytes(PreparedStmt, ColumnIndex);
if BlobSize > 0 then
   begin
   BlobStream := TMemoryStream(FRecordset.FMemoryStreams[FRecordset.AddMemoryStream(Buffer)]);
   BlobStream.Write(SqliteDbv3_column_blob(PreparedStmt, ColumnIndex)^, BlobSize);
   end
   else FRecordset.SetFieldNullValue(FRecordset.FRecordsCount, ColumnIndex , True);
end;


{ --- Navigating --- }

Procedure TSqlitePassRecordset.First;
begin
FCurrentRecordIndex := -1;
end;

Procedure TSqlitePassRecordset.Prior;
begin
MoveBy(-1)
end;

Procedure TSqlitePassRecordset.Next;
begin
MoveBy(1)
end;

Procedure TSqlitePassRecordset.Last;
begin
FCurrentRecordIndex := FRecordsCount;
{$IFDEF FPC}
//FOldActiveRecordIndex := FRecordsCount;
{$ENDIF}
end;

Procedure TSqlitePassRecordset.MoveBy(Offset: Integer);
begin
MoveState := grError;

Case Offset of
   { Move to Previous record }
   -1 : Dec(FCurrentRecordIndex);
    0 : ;
    1 : Inc(FCurrentRecordIndex);
    end;

if (FCurrentRecordIndex < 0)
   then begin
        MoveState := grBOF;
        FCurrentRecordIndex := -1;
        end
   else if (FCurrentRecordIndex > Pred(FRecordsCount))
           then begin
                MoveState := grEOF;
                FCurrentRecordIndex := FRecordsCount;
                end
           else begin
                MoveState := grOk;
                SetBookMarkFlag(FCurrentRecordIndex, bfCurrent);
                end;
end;


{ --- Managing Records (Create a new record, Add, Insert, Delete from records list) --- }

 { Set default values if needed }
(*For i := 0 to Pred(Fields.Count) do
    if Fields[i].FieldKind = fkData
       then Result^.FieldsValues[Pred(Fields[i].FieldNo)] := Fields[i].DefaultExpression;*)

procedure TSqlitePassRecordset.SetFRecordsCapacity(NewCapacity: Integer);
begin
  ReallocMem(FRecords, NewCapacity * SizeOf(PRecBuffer));
  FRecordsCapacity := NewCapacity;
end;

{ Add a new record to the end of the Records list and
  create a new memory bloc of records if necessary}
procedure TSqlitePassRecordset._AddRecord;
var
Delta: Integer;

begin
  { Get another bloc of memory if needed }
  if FRecordsUsedInMemoryBloc = FRecordMemoryBlocCapacity then
     begin
     FRecordsMemoryBlocs.Add(AllocMem(FRecordSize * FRecordMemoryBlocCapacity));
     FRecordsUsedInMemoryBloc := 0;
     end;
  { Upsize the list if needed }
  if FRecordsCount = FRecordsCapacity then
     begin
     if FRecordsCapacity > 64 then Delta := FRecordsCapacity div 4
        else if FRecordsCapacity > 8 then Delta := 16
                else Delta := 4;
     SetFRecordsCapacity(FRecordsCapacity + Delta);
     end;
  { Assign the new record on the Memory Bloc Address }   
  FRecords[FRecordsCount] := PRecBuffer(FRecordsMemoryBlocs[Pred(FRecordsMemoryBlocs.Count)]) + (FRecordSize * FRecordsUsedInMemoryBloc);
end;

{ Add a new 'temporary' record
  AddTempRecord checks first for any recycled records memory bloc to reuse
  and upsize the memory if necessary
  Returns the index of the next available record
  We don't inc FRecordCount so Dataset 'doesn't know' about the new record
  until the PostRecord procedure is called  }
function TSqlitePassRecordset.AddTempRecord: Integer;
begin
Result := FRecordsCount;
if FRecycledRecords.Count > 0
   then FRecords^[Result] := PRecBuffer(FRecycledRecords[Pred(FRecycledRecords.Count)])
   else _AddRecord;

{ Initialize the new record with 'blank' values }
Move(FBlankRecordInitialization^, FRecords[Result]^, FRecordSize);
end;

{ Add a new record }
function TSqlitePassRecordset.AddRecord: Integer;
begin
  Result := AddTempRecord;
  PostRecord;
end;

{ Insert a new record :
  The record itself is added to the next available place
  in the last memory bloc.
  Then a new pointer is assigned to this record and inserted at the right place in the FRecords array }
Procedure TSqlitePassRecordset.InsertRecord(Index: Integer);
var
TempRecord: PRecBuffer;
begin
 TempRecord := FRecords[FRecordsCount];
 System.Move(FRecords[Index], FRecords[Index + 1],(FRecordsCount - Index) * SizeOf(PRecBuffer));
 FRecords[Index] := TempRecord;
end;

{ Make the record added with AddTempRecord 'visible' for the dataset }
procedure TSqlitePassRecordset.PostRecord;
begin
{ Go back to the last active record position }
If (FOldActiveRecordIndex > -1) then
   begin
   InsertRecord(FOldActiveRecordIndex);
   FCurrentRecordIndex := FOldActiveRecordIndex;
   FOldActiveRecordIndex := -1;
   end;

{ Remove the temp record from the Recycled list }
if FRecycledRecords.Count > 0
   then FRecycledRecords.Delete(Pred(FRecycledRecords.Count));

{ Makes the new record visible to the dataset.GetRecord function }
Inc(FRecordsCount);
Inc(FRecordsUsedInMemoryBloc);
end;

procedure TSqlitePassRecordset.Cancel;
begin
{ Move back to the Old selected record }
if FOldActiveRecordIndex > -1 then
   begin
   FCurrentRecordIndex := FOldActiveRecordIndex;
   FOldActiveRecordIndex := -1;
   end;
end;

{ Add a String and Returns the Index of the new String in FAnsiStrings Array }
Procedure TSqlitePassRecordset.AddAnsiString(Buffer: PRecBuffer; StrValue: AnsiString);
var
StrIndex,
LastRecycledString: Integer;
begin
 if FRecycledAnsiStrings.Count > 0
    then begin
         LastRecycledString := Pred(FRecycledAnsiStrings.Count);
         StrIndex := Integer(FRecycledAnsiStrings[LastRecycledString]);
         FAnsiStrings[StrIndex] := StrValue;
         FRecycledAnsiStrings.Delete(LastRecycledString);
         end
    else begin
         StrIndex := FAnsiStrings.Count;
         FAnsiStrings.Add(StrValue);
         end;

 Move(StrIndex, Buffer^, SizeOf(Integer));
end;

procedure TSqlitePassRecordset.AddAnsiString(RecordIndex, FieldIndex: Integer; StrValue: AnsiString);
var
Buffer: PRecBuffer;
begin
 Buffer := GetFieldBuffer(RecordIndex, FieldIndex);
 AddAnsiString(Buffer, StrValue);
 SetFieldNullValue(RecordIndex, FieldIndex, False);
end;

procedure TSqlitePassRecordset.ClearAnsiString(RecordIndex, FieldIndex: Integer);
var
StrIndex: Integer;
begin
if Not FieldIsNull(RecordIndex, FieldIndex) then
   begin
   StrIndex := PInteger((FRecords[RecordIndex] + FInternalFieldsOffset[FieldIndex]))^;
   FAnsiStrings[StrIndex] := '';
   FRecycledAnsiStrings.Add(PInteger(StrIndex));
   SetFieldNullValue(RecordIndex, FieldIndex, True);
   end;
end;


{ Add a WideString and Returns the Index of the new WideString in FWideStrings Array }
procedure TSqlitePassRecordset.AddWideString(Buffer: PRecBuffer; StrValue: UTF16WideString);
var
StrIndex,
LastRecycledWideString: Integer;
begin
 if FRecycledWideStrings.Count > 0
    then begin
         LastRecycledWideString := Pred(FRecycledWideStrings.Count);
         StrIndex := Integer(FRecycledWideStrings[LastRecycledWideString]);
         FWideStrings[StrIndex] := StrValue;
         FRecycledWideStrings.Delete(LastRecycledWideString);
         end
    else begin
         StrIndex := FWideStrings.Count;
         FWideStrings.Add(StrValue);
         end;
 Move(StrIndex, Buffer^, SizeOf(Integer));
 end;

procedure TSqlitePassRecordset.AddWideString(RecordIndex, FieldIndex: Integer; StrValue: UTF16WideString);
var
Buffer: PRecBuffer;
begin
 Buffer := GetFieldBuffer(RecordIndex, FieldIndex);
 AddWideString(Buffer, StrValue);
 SetFieldNullValue(RecordIndex, FieldIndex, False);
end;

procedure TSqlitePassRecordset.ClearWideString(RecordIndex, FieldIndex: Integer);
var
StrIndex: Integer;
begin
if Not FieldIsNull(RecordIndex, FieldIndex) then
   begin
   StrIndex := PInteger((FRecords[RecordIndex] + FInternalFieldsOffset[FieldIndex]))^;
   FWideStrings[StrIndex] := '';
   FRecycledWideStrings.Add(PInteger(StrIndex));
   SetFieldNullValue(RecordIndex, FieldIndex, True);
   end;
end;

{ Add a MemoryStream and Returns the index of the MemoryStream in FMemoryStreams }
function TSqlitePassRecordset.AddMemoryStream(Buffer: PRecBuffer): Integer;
var
LastRecycledIndex: Integer;
begin
  if FRecycledMemoryStreams.Count > 0
     then begin
     LastRecycledIndex := Pred(FRecycledMemoryStreams.Count);
     Result := Integer(FRecycledMemoryStreams[LastRecycledIndex]);
     FRecycledMemoryStreams.Delete(LastRecycledIndex);
     end
     else Result := FMemoryStreams.Add(TMemoryStream.Create);
  Move(Result, Buffer^, SizeOf(Integer));
end;

function TSqlitePassRecordset.AddMemoryStream(RecordIndex, FieldIndex: Integer): Integer;
var
Buffer: PRecBuffer;
begin
 Buffer := GetFieldBuffer(RecordIndex, FieldIndex);
 Result := AddMemoryStream(Buffer);
 SetFieldNullValue(RecordIndex, FieldIndex, False);
end;

procedure TSqlitePassRecordset.ClearMemoryStream(Buffer: PRecBuffer; FieldIndex: Integer);
var
BlobIndex: Integer;
begin
if Not FieldIsNull(Buffer, FieldIndex) then
   begin
   BlobIndex := PInteger(Buffer + FInternalFieldsOffset[FieldIndex])^;
   TMemoryStream(FMemoryStreams[BlobIndex]).Clear;
   FRecycledMemoryStreams.Add(PInteger(BlobIndex));
   SetFieldNullValue(Buffer, FieldIndex, True);
   end;
end;

procedure TSqlitePassRecordset.ClearMemoryStream(RecordIndex, FieldIndex: Integer);
begin
   ClearMemoryStream(FRecords[RecordIndex], FieldIndex);
end;

Function TSqlitePassRecordset.GetFieldAsBlob(RecordIndex, FieldIndex: Integer): TMemoryStream;
var
BlobIndex: Integer;
begin
if FieldIsNull(RecordIndex, FieldIndex)
   then Result := nil
   else begin
        BlobIndex := PInteger((FRecords[RecordIndex] + FInternalFieldsOffset[FieldIndex]))^;
        Result := FMemoryStreams[BlobIndex];
        end;
end;


{ Delete FRecords[Index] from the FRecords array and downsize the array
  The record itself is not freed
  NO checking of array bounds }
procedure TSqlitePassRecordset.DeleteRecord(Index: Integer);
begin
  ClearRecord(Index);
  Dec(FRecordsCount);
  Dec(FRecordsUsedInMemoryBloc);
  if Index < FRecordsCount
     then System.Move(FRecords[Succ(Index)], FRecords[Index],(FRecordsCount - Index) * SizeOf(pRecBuffer));
  FRecords[FRecordsCount] := nil;
end;

{ Copy a record from Source to Dest }
procedure TSqlitePassRecordset.CopyRecord(Source, Dest: PRecBuffer);
var
i, FieldPos, SourceBlobStreamIndex, DestBlobStreamIndex: Integer;

//StrIndex: Integer;
//SourceWideString: UTF16WideString;
SourceBuffer: PRecBuffer;
begin
{ We should not need to copy record if FDataset is not assigned, so it is not
  implemented }
if Not Assigned(FDataset) then Exit;

Move(Source^, Dest^, FRecordSize);
For i := 0 to Pred(FDataset.Fields.Count) do
   begin
   Case FDataset.Fields[i].DataType of

       { Copy Blobs }
        ftBlob, ftMemo, ftGraphic, ftFmtMemo, ftTypedBinary, ftBytes, ftVarBytes
       {$IFDEF HasftWideMemo}, ftWideMemo {$ENDIF} :
         begin
         FieldPos := FDataset.Fields[i].Tag;
         if Not FieldIsNull(Source, FieldPos) then
            begin
            SourceBuffer := Source + FInternalFieldsOffset[FieldPos];
            SourceBlobStreamIndex := PInteger(SourceBuffer)^;
            if SourceBlobStreamIndex > -1 then
               begin
               DestBlobStreamIndex := AddMemoryStream(Dest + FInternalFieldsOffset[FieldPos]);
               TMemoryStream(FMemoryStreams[DestBlobStreamIndex]).LoadFromStream(TMemoryStream(FMemoryStreams[SourceBlobStreamIndex]));
               end;
            end;   
         end;

// TODO - Not Necessary ?
{ Copy WideStrings }
{      ftWideString, ftBytes, ftVarBytes, ftFixedChar, ftWideWideString:
         begin
         FieldPos := FDataset.Fields[i].Tag;
         StrIndex :=  PInteger(Source + FInternalFieldsOffset[FieldPos])^;
         SourceWideString := FWideStrings[StrIndex];
         AddWideString(Dest + FInternalFieldsOffset[FieldPos], SourceWideString);
         end; }

     end; { Case }
end; { for }
end;


procedure TSqlitePassRecordset.ClearRecord(Index: Integer);
var
i: Integer;
begin
For i := 0 to Pred(FDataset.Fields.Count) do
   begin
   Case FDataset.Fields[i].DataType of
        { Clear Blobs }
        ftBlob, ftMemo, ftGraphic, ftFmtMemo, ftTypedBinary, ftBytes, ftVarBytes
        {$IFDEF HasftWideMemo}, ftWideMemo {$ENDIF}
                              : ClearMemoryStream(Index, FDataset.Fields[i].Tag);
        { Clear Strings content }
        ftString, ftFixedChar : ClearAnsiString(Index, FDataset.Fields[i].Tag);
        ftWideString          : ClearWideString(Index, FDataset.Fields[i].Tag);
        end;
   end; { for }
{ Clear Buffer }
Move(FBlankRecordInitialization^, FRecords[Index]^, FRecordSize);
FRecycledRecords.Add(PRecBuffer(FRecords[Index]));
end;

Procedure TSqlitePassRecordset.ExchangeRecords(RecordIndex1, RecordIndex2: Integer);
var
TempRecord: PRecBuffer;
begin
 TempRecord := FRecords[RecordIndex1];
 FRecords[RecordIndex1] := FRecords[RecordIndex2];
 FRecords[RecordIndex2] := TempRecord;
end;


{ --- Active record old values backup / restore when editing --- }

procedure TSqlitePassRecordset.SaveActiveRecordOldValues;
begin
  CopyRecord(FRecords[ActiveRecordIndex], FActiveRecordOldValues);
end;

procedure TSqlitePassRecordset.RestoreActiveRecordOldValues;
begin
  CopyRecord(FActiveRecordOldValues, FRecords[ActiveRecordIndex]);
end;

procedure TSqlitePassRecordset.ClearActiveRecordOldValues;
var
i: Integer;
begin
For i := 0 to Pred(FDataset.Fields.Count) do
   Case FDataset.Fields[i].DataType of
       { Clear Blobs }
       ftBlob, ftMemo, ftGraphic, ftFmtMemo, ftTypedBinary, ftBytes, ftVarBytes
       {$IFDEF HasftWideMemo}, ftWideMemo {$ENDIF} : ClearMemoryStream(FActiveRecordOldValues, FDataset.Fields[i].Tag);
       end;
{ Initialize the new record with 'blank' values }
Move(FBlankRecordInitialization^, FActiveRecordOldValues^, FRecordSize);
end;


{ Read / Write Record BookMarkFlag }
Function TSqlitePassRecordset.GetBookMarkFlag(RecordIndex: Integer): TBookMarkFlag;
begin
  Result := PBookMarkFlag(FRecords[RecordIndex])^;
end;

Procedure TSqlitePassRecordset.SetBookMarkFlag(RecordIndex: Integer; Value: TBookMarkFlag);
begin
{$IFDEF FPC}
  Move(Value, FRecords[RecordIndex]^, SizeOf(Value));
{$ELSE}
  PBookMarkFlag(FRecords[RecordIndex])^ := Value;
{$ENDIF}
end;

{ Read / Write a record BookMarkData }
Function TSqlitePassRecordset.GetBookMarkData(RecordIndex: Integer): Cardinal;
begin
  Result := PCardinal(FRecords[RecordIndex]+ FBookMarkDataOffset)^;
  if FOldActiveRecordIndex = -1
     then FOldActiveRecordIndex := RecordIndex;
end;

Procedure TSqlitePassRecordset.SetBookMarkData(RecordIndex: Integer; Value: Cardinal);
begin
  PCardinal(FRecords[RecordIndex]+ FBookMarkDataOffset)^ := Value;
end;

{ Read / Write Record RecordState }
Function TSqlitePassRecordset.GetRecordState(RecordIndex: Integer):  TSqlitePassRecordState;
begin
  Result := TSqlitePassRecordState((FRecords[RecordIndex] + FRecordStateOffset)^)
end;

Procedure TSqlitePassRecordset.SetRecordState(RecordIndex: Integer; Value: TSqlitePassRecordState);
var
PreviousRecordState: TSqlitePassRecordState;

begin
if FPostPoned then
  begin
  PreviousRecordState := GetRecordState(RecordIndex);
  Case PreviousRecordState of
       { rsUnModified can be changed to any other state }
       rsUnModified : begin
                      //TSqlitePassRecordState((FRecords[RecordIndex] + FRecordStateOffset)^) := Value;
                      if Value = rsUnModified
                         then Move(FUnModifiedRecordInitialization^,
                                   (FRecords[RecordIndex] + FFieldsModifiedValuesOffset)^,
                                   FFieldsCount);
                      end;
       { rsModified can be changed only to rsDeleted }
       rsModified   : if Value = rsDeleted
                         then ;//TSqlitePassRecordState((FRecords[RecordIndex] + FRecordStateOffset)^) := Value;

       { rsModified can be changed only to rsDeleted. In this case, record is deleted from Memory }
       rsInserted   : if Value = rsDeleted
                         then DeleteRecord(RecordIndex);
       end;
       { rsDeleted  cannot be changed to another value }
  end;
end;

{ Read / Write Record Rowid - TableNo is zero indexed }
Function TSqlitePassRecordset.GetRowId(RecordIndex: Integer; TableNo: Word): Int64;
begin
  if FRowidTablesCount > 0
     then Result := PInt64(FRecords[RecordIndex] + FRowIdsOffset + (TableNo * SizeOf(Int64)))^
     else Result := -1;
end;

Procedure TSqlitePassRecordset.SetRowId(RecordIndex: Integer; TableNo: Word; Value: Int64);
begin
  PInt64(FRecords[RecordIndex] + FRowIdsOffset + (TableNo * SizeOf(Int64)))^ := Value;
end;


{ FieldIndex is zero indexed }
Function TSqlitePassRecordset.GetFieldBuffer(RecordIndex, FieldIndex: Integer): PRecBuffer;
begin
  Result := FRecords[RecordIndex] + FInternalFieldsOffset[FieldIndex];
end;

{ Strings }
Function TSqlitePassRecordset.GetFieldAsAnsiString(RecordIndex, FieldIndex: Integer): AnsiString;
var
StrIndex: Integer;
begin
if FieldIsNull(RecordIndex, FieldIndex)
   then Result := ''
   else begin
        StrIndex := PInteger((FRecords[RecordIndex] + FInternalFieldsOffset[FieldIndex]))^;
        Result := FAnsiStrings[StrIndex];
        end;
end;

Procedure TSqlitePassRecordset.SetFieldAsAnsiString(RecordIndex, FieldIndex: Integer; Value: AnsiString);
var
StrIndex: Integer;
begin
if FieldIsNull(RecordIndex, FieldIndex)
   then AddAnsiString(RecordIndex, FieldIndex, Value)
   else begin
        StrIndex :=  PInteger((FRecords[RecordIndex] + FInternalFieldsOffset[FieldIndex]))^;
        FAnsiStrings[StrIndex] := Value;
        end;
end;

{ WideStrings }
Function TSqlitePassRecordset.GetFieldAsWideString(RecordIndex, FieldIndex: Integer): UTF16WideString;
var
StrIndex: Integer;
begin
if FieldIsNull(RecordIndex, FieldIndex)
   then Result := ''
   else begin
        StrIndex := PInteger((FRecords[RecordIndex] + FInternalFieldsOffset[FieldIndex]))^;
        Result := FWideStrings[StrIndex];
        end;
end;

Procedure TSqlitePassRecordset.SetFieldAsWideString(RecordIndex, FieldIndex: Integer; Value: UTF16WideString);
var
StrIndex: Integer;
begin
if FieldIsNull(RecordIndex, FieldIndex)
   then AddWideString(RecordIndex, FieldIndex, Value)
   else begin
        StrIndex :=  PInteger((FRecords[RecordIndex] + FInternalFieldsOffset[FieldIndex]))^;
        FWideStrings[StrIndex] := Value;
        end;
end;

{ Variants }
Function TSqlitePassRecordset.GetFieldAsVariant(RecordIndex, FieldIndex: Integer; Field: TField): Variant;
var
Buffer: PRecBuffer;
begin
Result := UnAssigned;
Buffer := GetFieldBuffer(RecordIndex, FieldIndex);

if FieldIsNull(RecordIndex, FieldIndex)
   then Result := Null
   else
   Case Field.DataType of
        ftWord,
        ftSmallint,
        ftInteger,
        ftDate,
        ftTime      : Result := PInteger(Buffer)^;
        ftBoolean   : Result := PWordBool(Buffer)^;
        ftFloat,
        ftCurrency,
        ftBCD       : Result := PDouble(Buffer)^;
        ftDateTime  : Result := VarFromDateTime(PDouble(Buffer)^);
        ftString, ftBytes
                    : Result := GetFieldAsAnsiString(RecordIndex, FieldIndex);
        ftWideString: Result := GetFieldAsWideString(RecordIndex, FieldIndex);

        end;
end;


{ Read and Write a record 'Null' Boolean Flag }
Function TSqlitePassRecordset.FieldIsNull(RecordIndex, FieldIndex: Integer): Boolean;
begin
  Result := FieldIsNull(FRecords[RecordIndex], FieldIndex);
end;

Function TSqlitePassRecordset.FieldIsNull(Buffer: PRecBuffer; FieldIndex: Integer): Boolean;
begin
  Result := PByteBool(Buffer + FFieldsNullValuesOffset + (SizeOf(ByteBool) * FieldIndex))^ = True;
end;

Procedure TSqlitePassRecordset.SetFieldNullValue(RecordIndex, FieldIndex: Integer; Value: ByteBool);
begin
  SetFieldNullValue(FRecords[RecordIndex], FieldIndex, Value);
end;

Procedure TSqlitePassRecordset.SetFieldNullValue(Buffer: PRecBuffer; FieldIndex: Integer; Value: ByteBool);
begin
  PByteBool(Buffer + FFieldsNullValuesOffset + (SizeOf(ByteBool) * FieldIndex))^ := Value;
end;

{ Read and Write a record 'Modified' Boolean Flag }
Function TSqlitePassRecordset.FieldWasModified(RecordIndex, FieldIndex: Integer): Boolean;
begin
  Result := PByteBool(FRecords[RecordIndex] + FFieldsModifiedValuesOffset + (SizeOf(ByteBool) * FieldIndex))^ = True;
end;

Procedure TSqlitePassRecordset.SetFieldWasModified(RecordIndex, FieldIndex: Integer; Value: ByteBool);
begin
  PByteBool(FRecords[RecordIndex] + FFieldsModifiedValuesOffset + (SizeOf(ByteBool) * FieldIndex))^ := Value;
end;

{ --- Direct access to fields values, when no TFields are defined ---
  RecordIndex and FieldIndex are zero based }

{ Returns a String for a given RecordIndex and FieldIndex }
Function TSqlitePassRecordset.GetTextItem(RecordIndex, FieldIndex: Integer): String;
var
StrIndex: Integer;
begin
if FieldIsNull(RecordIndex, FieldIndex)
   then Result := ''
   else begin
        StrIndex := PInteger(FRecords[RecordIndex] + FFieldsValuesOffset + (SizeOf(Integer)* FieldIndex))^;
        if FDatabase.SystemEncoding = sysUTF8
           then Result := FAnsiStrings[StrIndex]
           else Result := FWideStrings[StrIndex];
        end;
end;

{ Set a String (Ansi or WideString) for a given RecordIndex and FieldIndex }
Procedure TSqlitePassRecordset.SetTextItem(RecordIndex, FieldIndex: Integer; Value: String);
var
StrIndex: Integer;
begin
  StrIndex := PInteger(FRecords[RecordIndex] + FFieldsValuesOffset + (SizeOf(Integer)* FieldIndex))^;
  if FDatabase.SystemEncoding = sysUTF8
     then FAnsiStrings[StrIndex] := Value
     else FWideStrings[StrIndex] := Value;
end;

{ Add a String (Ansi or WideString) for a given RecordIndex and FieldIndex
  Field Index is zero Indexed }
Procedure TSqlitePassRecordset.AddTextItem(RecordIndex, FieldIndex: Integer; Value: String);
var
Buffer: PRecBuffer;
begin
  Buffer := PRecBuffer(FRecords[RecordIndex] + FFieldsValuesOffset + (SizeOf(Integer)* FieldIndex));
  if FDatabase.SystemEncoding = sysUTF8
     then AddAnsiString(Buffer, Value)
     else AddWideString(Buffer, Value);
  SetFieldNullValue(RecordIndex, FieldIndex, False);
end;


{ Internal information to retrieve a internal field position or a internal field offset
  in buffer from a given TDataset.Field.Index }
procedure TSqlitePassRecordset.InitInternalFieldsInfos(Fields: TFields);
var
i, j: Integer;
begin
FDataset.FHasCalcFields := False;
j := 0;
for i := 0 to Pred(Fields.Count) do
    begin
    Case Fields[i].FieldKind of
         fkData:      Fields[i].Tag := Pred(Fields[i].FieldNo);

         fkCalculated,
         fkLookup:    begin
                      Fields[i].Tag := FDataFieldsCount + j;
                      Inc(j);
                      FDataset.FHasCalcFields := True;
                      end;
        end; { Case }
    end;
end;

{ Create a sublist of active (filtered) Dataset.Filter to speed up filters operations }
procedure TSqlitePassRecordset.InitInternalFilters(Fields: TFields);
var
i: Integer;
FFieldFilter: TSqlitePassFieldFilter;
begin
FDataset.FFieldFiltersFiltered.Clear;
for i := 0 to Pred(Fields.Count) do
    begin
    FFieldFilter := FDataset.FFieldFilters[Fields[i].Tag];
    if FFieldFilter.Filtered
       then FDataset.FFieldFiltersFiltered.Add(FFieldFilter);
    end;
end;

{ Set all the Calculated and Lookup fields as Null }
procedure TSqlitePassRecordset.ClearCalcFields(RecordIndex: Integer);
var
CalcFieldsNullValues: PRecBuffer;
begin
if Assigned(FCalcFieldsNullValuesInitBuffer) then
   begin
   CalcFieldsNullValues := FRecords[RecordIndex] + FFieldsNullValuesOffset + FDataFieldsCount;
   Move(FCalcFieldsNullValuesInitBuffer, CalcFieldsNullValues, FCalcFieldsCount);
   end;
end;

{ Sorting }

Procedure TSqlitePassRecordset.SortRecords(SortedFields: TList);

  function CompareRecords(RecordIndex1: Integer; RecordBuffer: PRecBuffer): Integer;
  var
  i: Integer;
  begin
  Result := 0;
  for i := 0 to Pred(SortedFields.Count) do
      begin
      Result := CompareFieldValues(TSqlitePassSortedField(SortedFields[i]).FieldIndex,
                                   RecordIndex1, RecordBuffer,
                                   TSqlitePassSortedField(SortedFields[i]).Descendant);
      { Skips the others fields comparisons if values are already different }
      if Result <> 0 then Break
      end;
  end;

  Procedure InternalSortRecords(iLo, iHi: Integer);
  var
    Lo, Hi: Integer;
    MidRecordBuffer: PPRecBuffer;
  begin
    Lo := iLo;
    Hi := iHi;
    New(MidRecordBuffer);
    MidRecordBuffer^ := FRecords[(Lo + Hi) shr 1];
    repeat
     while CompareRecords(Lo, MidRecordBuffer^) < 0 do Inc(Lo);
     while CompareRecords(Hi, MidRecordBuffer^) > 0 do Dec(Hi);
     if Lo <= Hi then
        begin
        ExchangeRecords(Lo,Hi);
        Inc(Lo);
        Dec(Hi);
        end;
    until Lo > Hi;
    if Hi > iLo then InternalSortRecords(iLo, Hi);
    if Lo < iHi then InternalSortRecords(Lo, iHi);
    Dispose(MidRecordBuffer);
  end;
begin { SortRecords }
  InternalSortRecords(0, Pred(FRecordsCount));
end;


{ Utility function to compare two values based on the same DataType.
  This function is used to sort the recordset }
function TSqlitePassRecordset.CompareFieldValues(Const FieldIndex: Integer; RecordIndex1: Integer; RecordBuffer: PRecBuffer; Descending: Boolean = False): Integer;
var
FieldPos: Integer;
Buffer1, Buffer2: PRecBuffer;
Buffer1IsNull, Buffer2IsNull: Boolean;
MemoryStream1, MemoryStream2: TMemorySTream;
MemoText1, MemoText2: UTF16WideString;
DataType: TFieldType;
begin
Result := 0;
FieldPos := FDataset.Fields[FieldIndex].Tag;
Buffer1IsNull := FieldIsNull(RecordIndex1, FieldPos);
Buffer2IsNull := (PByteBool(RecordBuffer + FFieldsNullValuesOffset + FieldPos)^ = True);

if Buffer1IsNull and Not Buffer2IsNull then Result := -1
   else if Buffer2IsNull and Not Buffer1IsNull then Result := 1
        { The two values are NOT Null }
        else if Not (Buffer1IsNull and Buffer2IsNull) then
                begin
                Buffer1 := GetFieldBuffer(RecordIndex1, FieldPos);
                Buffer2 := RecordBuffer + FInternalFieldsOffset[FieldPos];
                DataType := FDataset.Fields[FieldIndex].DataType;

                Case DataType of
                      ftBoolean :
                      if PWordBool(Buffer1)^ > PWordBool(Buffer2)^ then Result := 1
                         else if PWordBool(Buffer1)^ < PWordBool(Buffer2)^ then Result := -1;

                      ftWord, ftSmallint, ftInteger, ftDate, ftTime :
                      if PInteger(Buffer1)^ > PInteger(Buffer2)^ then Result := 1
                         else if PInteger(Buffer1)^ < PInteger(Buffer2)^ then Result := -1;

                      ftAutoInc, ftLargeInt :
                      if PInt64(Buffer1)^ > PInt64(Buffer2)^ then Result := 1
                         else if PInt64(Buffer1)^ < PInt64(Buffer2)^ then Result := -1;

                      ftBCD, ftCurrency, ftFloat, ftDateTime :
                      if PDouble(Buffer1)^ > PDouble(Buffer2)^ then Result := 1
                         else if PDouble(Buffer1)^ < PDouble(Buffer2)^ then Result := -1;

                      ftString, ftFixedChar :
                      Result := AnsiCompareStr(FAnsiStrings[PInteger(Buffer1)^], FAnsiStrings[PInteger(Buffer2)^]);

                      ftWideString :
                      Result := AnsiCompareStr(FWideStrings[PInteger(Buffer1)^], FWideStrings[PInteger(Buffer2)^]);

                      ftMemo, ftFmtMemo {$IFDEF HasftWideMemo}, ftWideMemo {$ENDIF}:
                      begin
                      MemoryStream1 := FMemoryStreams[PInteger(Buffer1)^];
                      MemoryStream2 := FMemoryStreams[PInteger(Buffer2)^];
                      if MemoryStream1.Size > 0
                         then SetString(MemoText1, pChar(MemoryStream1.Memory), MemoryStream1.Size)
                         else MemoText1 := '';

                      if MemoryStream2.Size > 0
                         then SetString(MemoText2, pChar(MemoryStream2.Memory), MemoryStream2.Size)
                         else MemoText2 := '';

                      Result := AnsiCompareStr(MemoText1, MemoText2);
                      end;
                   end; { Case }
                end; { if Not (Buffer1IsNull and Buffer2IsNull) }

{ Inverts the result if Descending Order }
If Descending and (Result <> 0)
   then Result := Result * -1;
end;

{ --- Clear Postponed Records Changes --- }
Procedure TSqlitePassRecordset.ClearModifiedRecords;
begin
  SetLength(FModifiedRecords, 0);
end;


